# -*- coding: utf-8 -*-
"""
This module implements function class, a key component for forward mode autodifferentiation.

"""
import numpy as np
from .dual_number import DualNumber

class function:
    '''
    Create a function object to handle evalutation of a function at particular x coordinates
    and compute corresponding Jacobian through forward mode automatic differentiation.

    '''

    def __init__(self, *f, x_dim=1):
        """
        Record user defined function.
		
        ----------
        f : user defined function(s)
            A function or multiple functions that takes in a vector of x or one x as a scalar and compute arithmetic result.
        x_dim : int
            Specify how many independent variables there are in all input functions.
	    
        """
        self.x_dim = x_dim
        self.function_list = f #list of functions each returning one output

    def _plugin(self,x): # x is an iterable
        """
        Process independent variables as a vector input.
		
        ----------
        x : int or float or array-like of numeric values.
            Specify how many independent variables there are in all input functions.
	    
        """
        results = []
        for f in self.function_list:
            try:
                result = f(*x)
            except TypeError:
                try:
                    result = f(x)
                except TypeError:
                    raise TypeError ('Function fails to process given input.')
            results.append(result)
        results = np.array(results)
        return results


    def __call__(self,*x):
        """
        Execute user defined function.
		
        Parameters
        ----------
        f : user defined function
            A function that takes in a vector or scalar of x and compute arithmetic result.
        x : array_like
            An array of numeric values (int or float).
            It could be a scalar (int or float).
			
        Returns
        -------
        f : user defined function
            A function evaluated at x. 
	    
        Notes
        ----------
        NumPy is known to return nan with integer inputs in cases
	when a mathematically correct answer should be a float number, 
        e.g. print([10**c for c in np.arange(-5,5)]) will raise a ValueError in Numpy
        saying ValueError: Integers to negative integer powers are not allowed.
        As such we convert all input to float to avoid such issues.
			
        """
        # Preprocess x accepting all kinds of input and make them into 1-d array
        x = np.array(x).reshape(-1).astype('float')  # convert numeric input to float to prevent NumPy delegated computation returning exact result for integer input
        if len(x) != self.x_dim:
            raise ValueError('Dimension Mismatch')
        result = self._plugin(x)
        if result.shape[0] == 1:
            return result[0]
        else: 
            return result

    def grad(self,*x):
        """
        Compute Jacobian based on user specified function(s), based on input independent variable values.
		
        Parameters
        ----------
        x : array_like
            An array of numeric values (int or float).
            It could be a scalar (int or float).
			
        Returns
        -------
        J : array_like
            Jacobian matrix for given function.
	    
        Examples
        --------
        >>> def f(x): # user-defined function
        >>>   return ad.sin(ad.exp(x)) + (6**ad.cos(x)/x)**x - 2*ad.tan(-x)
        >>> f = function(f) # initiate a function object
	>>> print("f(1) = {:.2f}".format(f(1))) # f(1) rounded to 2dp should be 6.16
        f(1) = 6.16
	>>> print("df/dx(1) = {:.2f}".format(f.grad(1))) # f'(1) rounded to 2dp should be 0.32
        df/dx(1) = 0.32
	
        """
        # Preprocess x
        x = np.array(x).reshape(-1) #accepting all kinds of input and make them into 1-d array
        if len(x) != self.x_dim:
            raise ValueError('Dimension Mismatch')
        
        J = []
        for i in range(self.x_dim):  
            p = np.identity(self.x_dim)[:,i].tolist()
            J.append(self._grad(x,p))
        result = np.array(J).T
        #putting into desired dimension, can be modified, right now we return 1-d array for R1 to Rn and Rm to R1
        if result.shape[0] == 1 and result.shape[1] ==1:
            return result[0,0]
        elif result.shape[0] == 1:
            return result[0,:]
        elif result.shape[1] == 1:
            return result[:,0]
        else: 
            return result

    def _grad(self,x,p):
        """
        Computes the derivative w.r.t. each input.
		
        Parameters
        ----------
        x : array_like
            An array of numeric values (int or float) as independent variables.
            It could be a scalar (int or float).
	p : array_like
            An array of numeric values (int or float) as seed directional vector.
            It could be a scalar (int or float).
			
        Returns
        -------
        J : array_like
            Jacobian matrix for given function.

        """
        dual_nums = [DualNumber(*input) for input in zip(x,p)]
        results = self._plugin(dual_nums)
        return np.array([result.dual for result in results])
        



# -*- coding: utf-8 -*-
"""

This module implements DualNumber class, a key component 
for forward mode automatic differentiation.

"""
import numpy as np

class DualNumber(object):
    '''
    Create a dual number

    Attributes
    ----------
    _supported_scalars : tuple
        Description of supported numerical types.    
    '''

    _supported_scalars = (int, float)
    
    def __init__(self,real,dual = None):
        """
        Specify the real and dual part of a dual number.

        Parameters
        ----------
        real : float
            Real part of dual number.
        dual : float
            Dual part of dual number, default to 1 if not specified.
        
        Note
        ----
        Dual part is typically only set to 1 through external function
        when finding derivative with respect to that dual number.
        
        """
        self.real = real
        if dual == None:
            self.dual = 1.0
        else:
            self.dual = dual

    def __repr__(self):
        """
        Nice string representation of dual number.

        Returns
        ----------
        str

        Examples
        --------
        Print DualNumber

        >>> x = DualNumber(1.0)
        >>> x
        DualNumber(real=1.0, dual=None)

        """
        args = f"real={repr(self.real)}, dual={repr(self.dual)}"
        return f"{type(self).__name__}({args})"

    def __add__(self, other):
        '''
        Compute addition with dual number or regular number.

        Parameters
        ----------
        other : int or float or DualNumber instance
            Other number being added.
        
        Returns
        ----------
        DualNumber
        
        Raises
        ------
        TypeError
        If the other number inputted is not of any supported numeric format. 
        
        '''
        if not isinstance(other, (*self._supported_scalars, DualNumber)):
            raise TypeError(f"Unsupported type '{type(other)}'")
        if isinstance(other, self._supported_scalars):
            return DualNumber(self.real+other, self.dual)
        else:
            return DualNumber(self.real + other.real, self.dual + other.dual)
    
    def __radd__(self, other):
        '''
        Compute reflective addition with regular number.

        Parameters
        ----------
        other : int or float
            Other number being added to.
        
        Returns
        ----------
        DualNumber
        
        '''
        return self.__add__(other)
    
    def __sub__(self, other):
        '''
        Compute Subtraction of dual number or regular number from dual number or regular number.

        Parameters
        ----------
        other : int or float or DualNumber instance
            Other number being subtracted.
        
        Returns
        ----------
        DualNumber
        
        Raises
        ------
        TypeError
        If the other is not of any supported numeric format.       
        
        '''
        if not isinstance(other, (*self._supported_scalars, DualNumber)):
            raise TypeError(f"Unsupported type '{type(other)}'")
        if isinstance(other, self._supported_scalars):
            return DualNumber(self.real - other, self.dual)
        else:
            return DualNumber(self.real - other.real, self.dual - other.dual)

    def __rsub__(self, other):
        '''
        Compute Subtraction of dual number from regular number.

        Parameters
        ----------
        other : int or float or DualNumber instance
            Other number being subtracted from.
        
        Returns
        ----------
        DualNumber
        
        Raises
        ------
        TypeError
        If the other is not of any supported numeric format.        
        
        '''
        if not isinstance(other, self._supported_scalars):
            raise TypeError(f"Unsupported type '{type(other)}'")
        else:
            return DualNumber(other - self. real, - self.dual)

    def __mul__(self, other):
        '''
        Compute Multiplication with dual number or regular number.

        Parameters
        ----------
        other : int or float or DualNumber instance
            Other number being multiplied.
        
        Returns
        -------
        DualNumber
        
        Raises
        ------
        TypeError
        If the other number inputted is not of any supported numeric format.      

        '''
        if not isinstance(other, (*self._supported_scalars, DualNumber)):
            raise TypeError(f"Unsupported type '{type(other)}'")
        if isinstance(other, self._supported_scalars):
            return DualNumber(self.real * other, self.dual * other)
        else:
            return DualNumber(self.real * other.real, self.real * other.dual + self.dual * other.real)
    
    def __rmul__(self, other):
        '''
        Compute reflective multiplication with regular number.

        Parameters
        ----------
        other : int or float
            Other number being multiplied.
        
        Returns
        -------
        DualNumber
        
        Raises
        ------
        TypeError
        If the other number input is not of any supported numeric format.    
        
        '''
        return self.__mul__(other)

    def __pow__(self,other):
        '''
        Compute power raised to input regular number or dual number.

        Parameters
        ----------
        other : int or float or DualNumber instance
            Other number as the power a dual number is being raised to.
        
        Returns
        -------
        DualNumber
        
        Raises
        ------
        TypeError
        If the other number input is not of any supported numeric format.   
        
        Notes
        ----------
        Since we delegate power to NumPy and NumPy return integer or nothing when power 
        is a negative integer when a mathematically correct answer should be a float number, 
        e.g. print([10**c for c in np.arange(-5,5)]) will raise a ValueError in Numpy
        saying ValueError: Integers to negative integer powers are not allowed.
        As such we convert the power numeric terms to float to avoid such issues.
        
        '''
        if not isinstance(other, (*self._supported_scalars, DualNumber)):
            raise TypeError(f"Unsupported type '{type(other)}'")
        elif isinstance(other, self._supported_scalars):
            return DualNumber(self.real**float(other), other*self.real**float(other-1)*self.dual)
        else:
            return DualNumber(self.real**float(other.real),self.real**float(other.real)*(other.real*self.dual/self.real+other.dual*np.log(self.real)))
    
    def __rpow__(self,other):
        '''
        Compute input int or float number raised to the power of dual number.

        Parameters
        ----------
        other : int or float
            Other number as the base to which is raised to the power of a dual number.
        
        Returns
        -------
        DualNumber
        
        Raises
        ------
        TypeError
        If the other number input is not of any supported numeric format.    
        
        '''
        if not isinstance(other, self._supported_scalars):
            raise TypeError(f"Unsupported type '{type(other)}'")
        other = DualNumber(other,0)
        return other.__pow__(self)
        
    def __neg__(self):
        '''
        Flip the sign of input dual number by negating its dual part and real part respectively.
        
        Returns
        -------
        DualNumber 
        
        '''
        return DualNumber(- self.real , - self.dual)
    
    def __eq__(self, other):
        '''
        Compare the real part of two numbers.

        Parameters
        ----------
        other: int or float or DualNumber instance
        
        Returns
        -------
        Boolean 
        
        '''
        if not isinstance(other, (*self._supported_scalars, DualNumber)):
            raise TypeError(f"Unsupported type '{type(other)}'")
        elif isinstance(other, self._supported_scalars):
            return self.real == other
        else:
            return self.real == other.real
    
    def __ne__(self, other):
        '''
        Compare the real part of two numbers.

        Parameters
        ----------
        other: int or float or DualNumber instance
        
        Returns
        -------
        Boolean 
        
        '''
        if not isinstance(other, (*self._supported_scalars, DualNumber)):
            raise TypeError(f"Unsupported type '{type(other)}'")
        elif isinstance(other, self._supported_scalars):
            return not self.real == other
        else:
            return not self.real == other.real
    
    def __lt__(self, other):
        '''
        Compare the real part of two numbers using "<"

        Parameters
        ----------
        other: int or float or DualNumber instance
        
        Returns
        -------
        Boolean 
        
        '''
        if not isinstance(other, (*self._supported_scalars, DualNumber)):
            raise TypeError(f"Unsupported type '{type(other)}'")
        elif isinstance(other, self._supported_scalars):
            return self.real < other
        else:
            return self.real < other.real
    
    def __le__(self, other):
        '''
        Compare the real part of two numbers using "<="

        Parameters
        ----------
        other: int or float or DualNumber instance
        
        Returns
        -------
        Boolean 
        
        '''
        if not isinstance(other, (*self._supported_scalars, DualNumber)):
            raise TypeError(f"Unsupported type '{type(other)}'")
        elif isinstance(other, self._supported_scalars):
            return self.real <= other
        else:
            return self.real <= other.real
    
    def __gt__(self, other):
        '''
        Compare the real part of two numbers using ">"

        Parameters
        ----------
        other: int or float or DualNumber instance
        
        Returns
        -------
        Boolean 
        
        '''
        if not isinstance(other, (*self._supported_scalars, DualNumber)):
            raise TypeError(f"Unsupported type '{type(other)}'")
        elif isinstance(other, self._supported_scalars):
            return self.real > other
        else:
            return self.real > other.real
    
    def __ge__(self, other):
        '''
        Compare the real part of two numbers using ">="

        Parameters
        ----------
        other: int or float or DualNumber instance
        
        Returns
        -------
        Boolean 
        
        '''
        if not isinstance(other, (*self._supported_scalars, DualNumber)):
            raise TypeError(f"Unsupported type '{type(other)}'")
        elif isinstance(other, self._supported_scalars):
            return self.real >= other
        else:
            return self.real >= other.real


    def __truediv__(self, other):
        '''
        Compute float division of dual number by int, float or dual number.

        Parameters
        ----------
        other : int or float or DualNumber instance
            Other number being divided.
        
        Returns
        -------
        DualNumber
        
        Raises
        ------
        ZeroDivisionError
            If the denominator other number's real part is zero.
            
        '''
        if not isinstance(other, (*self._supported_scalars, DualNumber)):
            raise TypeError(f"Unsupported type '{type(other)}'")
        elif isinstance(other, self._supported_scalars):
            return DualNumber(self.real/other, self.dual/other)
        else:
            return DualNumber(self.real/other.real, (self.dual*other.real - other.dual*self.real)/(other.real*other.real))

    def __rtruediv__(self, other):        
        '''
        Compute float division of real number by dual number.

        Parameters
        ----------
        other : int or float
            Other real number being divided.
        
        Returns
        -------
        DualNumber
        
        Raises
        ------
        ZeroDivisionError
            If the denominator dual number's real part is zero.
            
        '''
        if isinstance(other, self._supported_scalars):
            return DualNumber(other/self.real, (-1*other*self.dual)/(self.real*self.real))    
        else:
            raise TypeError(f"Unsupported type '{type(other)}'")

    #=================== inplace operation ===================#
    def __iadd__(self,other):
        '''
        Compute inplace addition of dual number by int, float or dual number.

        Parameters
        ----------
        other : int or float or DualNumber instance
            Other number being added.
        
        Returns
        -------
        DualNumber
        
        '''
        if not isinstance(other, (*self._supported_scalars, DualNumber)):
            raise TypeError(f"Unsupported type '{type(other)}'")
        elif isinstance(other, self._supported_scalars):
            self.real += other
            return self
        else:
            self.real += other.real
            self.dual += other.dual
            return self

    def __isub__(self,other):
        '''
        Compute inplace substraction of dual number by int, float or dual number.

        Parameters
        ----------
        other : int or float or DualNumber instance
            Other number being added.
        
        Returns
        -------
        DualNumber
        
        '''
        if not isinstance(other, (*self._supported_scalars, DualNumber)):
            raise TypeError(f"Unsupported type '{type(other)}'")
        elif isinstance(other, self._supported_scalars):
            self.real -= other
            return self
        else:
            self.real -= other.real
            self.dual -= other.dual
            return self
    
    def __imul__(self,other):
        '''
        Compute inplace multiplication of dual number with int, float or dual number.

        Parameters
        ----------
        other : int or float or DualNumber instance
            Other number being added.
        
        Returns
        -------
        DualNumber
        
        '''
        if not isinstance(other, (*self._supported_scalars, DualNumber)):
            raise TypeError(f"Unsupported type '{type(other)}'")
        elif isinstance(other, self._supported_scalars):
            self.real *= other
            self.dual *= other
            return self
        else:
            self.real,self.dual = self.real*other.real, self.dual*other.real + self.real*other.dual
            return self
    
    def __itruediv__(self,other):
        '''
        Compute inplace division of dual number by int, float or dual number.

        Parameters
        ----------
        other : int or float or DualNumber instance
            Other number being added.
        
        Returns
        -------
        DualNumber
        
        Raises
        ------
        ZeroDivisionError
            If the denominator other number's real part is zero.
            
        '''
        if not isinstance(other, (*self._supported_scalars, DualNumber)):
            raise TypeError(f"Unsupported type '{type(other)}'")
        elif isinstance(other, self._supported_scalars):
            self.real /= other
            self.dual /= other
            return self
        else:
            self.real,self.dual = self.real/other.real, (self.dual*other.real - self.real*other.dual)/(other.real*other.real)
            return self



from .operations import *
from .function import *
from .optimizers import *

__all__=['sin', 'cos', 'tan', 'sigmoid',  'arcsin', 'arccos', 'arctan', 'sinh', 'cosh', 'tanh', 'exp', 'log', 'sqrt', 'function',  'Adam', 'GD']




# -*- coding: utf-8 -*-
"""

This module implements different optimizers classes, which is the key extension functionality for finding local minimum and maximum.

"""
from abc import abstractmethod
import numpy as np
from .function import function


class Optimizer():
    '''
    Create a optimizer object to handle general optimization request
    for user defined learning rate, within maximum iterations and tolerance threshold.
    
    '''

    def __init__(self, learning_rate = 0.001, max_iteration = 10000, eps = 1e-15):
        """
        Initialize an optimizer.
		
        Parameters
        ----------
        learning_rate : float
            User specified step size at each iteration while moving towards the goal of the optimization task.
        
        max_iteration : int
            User specified maximum number of iterations a optimization task should perform.     
        
        eps : float
            User specified tolerance threshold. When an iteration's magnitude of change for independent variable  
            vector is smaller than the given threshold the optimization task is said to reach the goal and
            iteration would terminate and produce a final result.
            
        """
        if callable(learning_rate):
            self.eta = learning_rate
        else:
            self.eta = lambda t : learning_rate
        self.max_iteration = max_iteration
        self.eps = eps

    @abstractmethod
    def minimize(self,f, x_dim = 1, x0 = None,verbose=False):
        """
        A place-holder for minimize method that all children optimizer class must implement.
        
        """        
        raise NotImplementedError

class GD(Optimizer):
    """
    Gradient Optimizer Class used to find local minimum or maximum of a given function

    """
    def minimize(self,f, x_dim = 1, x0 = None,verbose=False):
        """
        Minimize a function using gradient descent. 
	Each iteration the algorithm make a step in the steepest descent.

        Parameters
        ----------
        f: function class instance
            The function to minimize.
        x_dim : int
            Specify how many independent variables there are in all input functions.
        x0: int or float or array-like of numeric values.
            The starting values of the independent variables for the gradietn descent algorithm
        verbose: bool
            Whether to return the full iteration history of the gradient descent algorithm

        Returns
        -------
        x: float or array-like. 
            A scalar x value for scalar input, or a list of x values for vector inputs.
	    This is the value(s) of independent variable(s) at the found local minimum.
        f(x): float
            The value of the function at the minimum
        history: list
            Each element of the list represents one iteration which is a tuple.
	    Each element of that iteration tuple is made of two elements: a NumPy array of x's and function value at that iteration.
	    The NumPy array of x's for a iteration consists of the corresponding value(s) of independent variable(s) in the input order.
        
        Examples
        --------
        >>> def f(x_1,x_2):
        >>>   return (x_1+1)**2 + x_2**2 + 3
        >>> gd = ad.GD(learning_rate = 0.01, max_iteration = 10000, eps = 1e-12)
        >>> id_vars, f_min = gd.minimize(f, x_dim=2, x0 = [12, 6])
	>>> x_1, x_2 = id_vars
        >>> print(f"the local maximum function value is {f_min:.2f}")
        the local maximum function value is 3.00
        >>> print(f"the corresponding independent variable values are x = {x_1:.2f}, y={x_2:.2f}")
        the corresponding independent variable values are x = -1.00, y=0.00
	
        """
        if isinstance(f,function):
            if len(f.function_list) > 1: # only functions with a single output can be minimized
                raise TypeError("Cannot optimize vector-valued function")
        else:
            f = function(f,x_dim = x_dim)
        if x0 is None:
            x0 = np.zeros(f.x_dim) # initialize the starting point to be zero (vector)
        x = x0
        t = 0
        history = []
        for i in range(self.max_iteration):
            x_new = x - self.eta(t)*f.grad(x) # update rule of gradient descent
            if abs(f(x)-f(x_new)) < self.eps: # break the loop when the change in function is less than eps
                x = x_new
                break
            x = x_new
            t += 1
            if verbose:
                history.append((x,f(x)))
	if len(x) == 1: x=x[0] # unpack scalar input into float number
        if verbose:
            return x,f(x),history
        else:
            return x,f(x)

    def maximize(self,f, x_dim = 1, x0 = None,verbose=False):
        """
        Maximize a function using gradient descent. 
	Each iteration the algorithm make a step in the steepest ascent.
	Search for the maximum of -f(x) which makes the iterative process equivalent. 

        Parameters
        ----------
        f: function class instance
            The function to maximize.
        x_dim : int
            Specify how many independent variables there are in all input functions.
        x0: int or float or array-like of numeric values.
            The starting values of the independent variables for the gradietn descent algorithm
        verbose: bool
            Whether to return the full iteration history of the gradient descent algorithm

        Returns
        -------
        x: float or array-like. 
            A scalar x value for scalar input, or a list of x values for vector inputs.
	    This is the value(s) of independent variable(s) at the found local maximum.
        f(x): float
            The value of the function at the maximum
        history: list
            Each element of the list represents one iteration which is a tuple.
	    Each element of that iteration tuple is made of two elements: a NumPy array of x's and function value at that iteration.
	    The NumPy array of x's for a iteration consists of the corresponding value(s) of independent variable(s) in the input order.
        
        Examples
        --------
        >>> def f(x,y):
        >>>   return -(x+2)*(x+1)-(y+1)*(y+3)
        >>> gd = ad.GD(learning_rate = 0.2, max_iteration = 6000, eps = 1e-6)
        >>> id_vars, f_max = gd.maximize(f, x_dim=2, x0 = [3.1, -2])
	>>> x, y = id_vars
        >>> print(f"the local maximum function value is {f_max:.2f}")
        the local maximum function value is 1.25
        >>> print(f"the corresponding independent variable values are x = {x:.2f}, y={y:.2f}")
        the corresponding independent variable values are x = -1.50, y=-2.00
	
        """
        if isinstance(f,function):
            if len(f.function_list) > 1: # only functions with a single output can be maximized
                raise TypeError("Cannot optimize vector-valued function")
            neg_f = function(lambda *x: -1*f.function_list[0](*x),x_dim=x_dim) # build function object neg_f whose "f" attribute equal to -f
        else:
            neg_f = function(lambda *x: -1*f(*x),x_dim = x_dim) # build function object neg_f whose "f" attribute equal to -f
        if verbose:
            x,neg_f_min,history = self.minimize(neg_f,x_dim=x_dim,x0=x0,verbose=True) # recall max f = - min(-f); minimizing neg_f gives -max(f) and argmax(f) 
            history = [(tup[0],-1*tup[1]) for tup in history] # correct history which evaluated -f(x)
            return x,-1*neg_f_min,history
        x,neg_f_min = self.minimize(neg_f,x_dim=x_dim,x0=x0,verbose=False)
        return x,-1*neg_f_min

class Adam(Optimizer):
    def __init__(self, learning_rate=0.001, max_iteration = 10000, beta_1 = 0.9, beta_2 = 0.999, epsilon = 1e-08):
        super().__init__(learning_rate, max_iteration)
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.epsilon = epsilon

    def minimize(self,f, x_dim = 1, x0 = None,verbose=False):
        if isinstance(f,function):
            if len(f.function_list) > 1: # only functions with a single output can be minimized
                raise TypeError("Cannot optimize vector-valued function")
        else:
            f = function(f,x_dim = x_dim)
        if x0 is None:
            x0 = np.zeros(f.x_dim) # initialize the starting point to be zero (vector)
        x = x0
        t = 0
        m = 0
        v = 0
        history = []
        for i in range(self.max_iteration):
            g = f.grad(x)
            m = self.beta_1 * m + (1-self.beta_1)*g
            v = self.beta_2 * v + (1-self.beta_2)*g**2
            m_hat = m/(1-self.beta_1**(t+1))
            v_hat = v/(1-self.beta_2**(t+1))
            x_new = x - self.eta(t)/np.sqrt(v_hat+self.epsilon)*m_hat # update rule of adam
            if abs(f(x)-f(x_new)) < self.eps: # break the loop when the change in function is less than eps
                x = x_new
                break
            x = x_new
            t += 1
            if verbose:
                    history.append((x,f(x)))
	if len(x) == 1: x=x[0] # unpack scalar input into float number
        if verbose:
            return x,f(x), history
        else:
            return x,f(x)
    
    def maximize(self,f, x_dim = 1, x0 = None,verbose=False):
        if isinstance(f,function):
            if len(f.function_list) > 1: # only functions with a single output can be maximized
                raise TypeError("Cannot optimize vector-valued function")
            neg_f = function(lambda *x: -1*f.function_list[0](*x),x_dim=x_dim) # build function object neg_f whose "f" attribute equal to -f
        else:
            neg_f = function(lambda *x: -1*f(*x),x_dim = x_dim) # build function object neg_f whose "f" attribute equal to -f
        if verbose:
            x,neg_f_min,history = self.minimize(neg_f,x_dim=x_dim,x0=x0,verbose=True) # recall max f = - min(-f); minimizing neg_f gives -max(f) and argmax(f) 
            history = [(tup[0],-1*tup[1]) for tup in history] # correct history which evaluated -f(x)
            return x,-1*neg_f_min,history
        x,neg_f_min = self.minimize(neg_f,x_dim=x_dim,x0=x0,verbose=False)
        return x,-1*neg_f_min





# -*- coding: utf-8 -*-
"""

This module implement overloading functions to handle arithmetic for dual numbers.

"""

import numpy as np
from .dual_number import DualNumber

def exp(x):
  '''
  Overloads the exponential function. 

  Parameters
  ----------
  x : int or float or DualNumber instance

  Returns
  -------
  float or DualNumber instance

  '''
  if isinstance(x,DualNumber):
    return DualNumber(np.exp(x.real), np.exp(x.real)*x.dual)
  else:
    return np.exp(x)

def _log(x):
  '''
  Computes the natural logarithm. 

  Parameters
  ----------
  x : int or float or DualNumber instance

  Returns
  -------
  float or DualNumber instance

  '''
  if isinstance(x,DualNumber):
    return DualNumber(np.log(x.real), x.dual/x.real)
  else:
    return np.log(x)

def log(x,base=None):
  '''
  Overloads the logarithm function for any base

  Parameters
  ----------
  x : int or float or DualNumber instance
  base: int or float or DualNumber instance

  Returns
  -------
  float or DualNumber instance

  '''
  if base is None:
    return _log(x)
  return _log(x)/_log(base)

def sqrt(x):
  '''
  Overloads the non-negative square-root function. 

  Parameters
  ----------
  x : int or float or DualNumber instance

  Returns
  -------
  float or DualNumber instance

  '''
  if isinstance(x,DualNumber):
    return DualNumber(np.sqrt(x.real),.5*x.real**-.5*x.dual)
  else:
    return np.sqrt(x)

def sigmoid(x):
  '''
  Computes the sigmoid function evaluated on x. 

  Parameters
  ----------
  x : int or float or DualNumber instance

  Returns
  -------
  float or DualNumber instance

  '''
  return 1/(1+exp(-x))

#=================== Trigonometric ===================#
def sin(x):
  '''
  Overloads the sin() function. 

  Parameters
  ----------
  x : int or float or DualNumber instance
      Each number represent angle in radians (:math:`2 \pi` rad equals 360 degrees).

  Returns
  -------
  float or DualNumber instance
      
  Examples
  --------
  Print sine of one angle:

  >>> ad.sin(np.pi/2.)
  1.0

  >>> x = DualNumber(np.pi/2)
  >>> ad.sin(x)
  DualNumber(real = 1., dual = 0.)
  '''
  if isinstance(x,DualNumber):
    return DualNumber(np.sin(x.real), np.cos(x.real)*x.dual)
  else:
    return np.sin(x)

def cos(x):
  '''
  Overloads the cos() function. 

  Parameters
  ----------
  x : int or float or DualNumber instance
      Each number represent angle in radians (:math:`2 \pi` rad equals 360 degrees).

  Returns
  -------
  float or DualNumber instance
      
  Examples
  --------
  Print cosine of one angle:

  >>> ad.cos(np.pi/2.)
  0.0

  >>> x = DualNumber(np.pi/2)
  >>> ad.cos(x)
  DualNumber(real = 0., dual = -1.)
  '''
  if isinstance(x,DualNumber):
    return DualNumber(np.cos(x.real), -1*np.sin(x.real)*x.dual)
  else:
    return np.cos(x)

def tan(x):  
  '''
  Overloads the tangent function. 

  Parameters
  ----------
  x : int or float or DualNumber instance

  Returns
  -------
  float or DualNumber instance
      
  '''
  if isinstance(x,DualNumber):
    return DualNumber(np.tan(x.real), x.dual/(np.cos(x.real)**2))
  else:
    return np.tan(x)
  
def arcsin(x):
  '''
  Overloads the inverse sine function. 

  Parameters
  ----------
  x : int or float or DualNumber instance

  Returns
  -------
  float or DualNumber instance
      
  '''
  if isinstance(x,DualNumber):
    return DualNumber(np.arcsin(x.real),x.dual/np.sqrt(1-x.real**2))
  else:
    return np.arcsin(x)

def arccos(x):
  '''
  Overloads the inverse cosine function. 

  Parameters
  ----------
  x : int or float or DualNumber instance

  Returns
  -------
  float or DualNumber instance
      
  '''
  if isinstance(x,DualNumber):
    return DualNumber(np.arccos(x.real),-1*x.dual/np.sqrt(1-x.real**2))
  else:
    return np.arccos(x)

def arctan(x):
  '''
  Overloads the inverse tangent function. 

  Parameters
  ----------
  x : int or float or DualNumber instance

  Returns
  -------
  float or DualNumber instance
      
  '''
  if isinstance(x,DualNumber):
    return DualNumber(np.arctan(x.real),x.dual/(1+x.real**2))
  else:
    return np.arctan(x)

#=================== Hyperbolic ===================#
def sinh(x):
  '''
  Overloads the hyperbolic sine function. 

  Parameters
  ----------
  x : int or float or DualNumber instance

  Returns
  -------
  float or DualNumber instance

  '''
  if isinstance(x,DualNumber):
    return DualNumber(np.sinh(x.real),x.dual*np.cosh(x.real))
  else:
    return np.sinh(x)  

def cosh(x):
  '''
  Overloads the hyperbolic cosine function. 

  Parameters
  ----------
  x : int or float or DualNumber instance

  Returns
  -------
  float or DualNumber instance

  '''
  if isinstance(x,DualNumber):
    return DualNumber(np.cosh(x.real),x.dual*np.sinh(x.real))
  else:
    return np.cosh(x) 

def tanh(x):
  '''
  Overloads the hyperbolic tangent function. 

  Parameters
  ----------
  x : int or float or DualNumber instance

  Returns
  -------
  float or DualNumber instance
      
  '''
  if isinstance(x,DualNumber):
    return DualNumber(np.tanh(x.real),x.dual/np.cosh(x.real)**2)
  else:
    return np.tanh(x) 


